
package expressions;

/*
	Env.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	Environments (= contexts).
*/

public class Env extends LinkList<Frame> implements Expr {

	/*
		The empty context.
	*/
	public final static Env EMPTY = new Env();

	/*
		Creates an empty environment.
	*/
	private Env() {
		super();
	}

	/*
		Creates a nonempty environment.
	*/
	public Env(Frame f, Env e) {
		super(f, e);
	}

	/*
		Creates an environment with one frame.
	*/
	public Env(Frame f) {
		this(f, EMPTY);
	}

	/*
		Aliasname for head.
	*/
	public Frame car() {
		return head();
	}

	/*
		Typesafe access to tail.
	*/
	public Env cdr() {
		return (Env) tail();
	}

	/*
		Binds sym to val.
	*/
	public void bind(Symbol sym, Expr val) throws Alarm {
		frameFor(sym).bind(sym, val);
	}

	/*
		Locally binds sym to val in the first frame.
	*/
	public void put(Symbol sym, Expr val) {
		car().put(sym, val);
	}

	/*
		Removes symbol sym from the innermost frame.
	*/
	public void delete(Symbol sym) throws Alarm {
		frameFor(sym).delete(sym);
	}

	/*
		Retrieves the value bound to sym, or returns null.
	*/
	public Expr retrieve(Symbol sym) {
		return frameFor(sym).retrieve(sym);
	}

	/*
		frameFor(sym) returns the innermost frame containing a
		binding for sym, if one exists, or the initial frame.
	*/
	private Frame frameFor(Symbol sym) {
		Env env = this;
		Frame frame = car();
		while (!env.cdr().isEmpty()) {
			if (frame.containsKey(sym)) break;
			env = env.cdr();
			frame = env.car();
		}
		return frame;
	}

	/*
		protects Symbol sym.
	*/
	public void protect(Symbol sym) {
		frameFor(sym).protect(sym);
	}

	/*
		Returns true if this Env protects sym.
	*/	
	public boolean protects(Symbol sym) {
		return frameFor(sym).protects(sym);
	}

	/*
		Evaluation.
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}

	/*
		Runtime type.
	*/
	public Expr type() {
		return Symbol.forName("Context");
	}

	/*
		Checks e for being an Env.
	*/
	public static Env checkEnv(Expr e, Expr client) throws Alarm {
		if (e instanceof Env) return (Env) e;
		throw new Alarm("ENV", client + " expects a context");
	}

	/*
		Returns a new global environment.
	*/
	public static Env globalEnv() {
		Env env = new Env(new Frame());
		try {
			Symbol op = Symbol.forName("op");
			env.put(op, Operator.forName("op"));
			env.protect(op);
		} catch (Alarm alarm) {
			// Thrown when op[op] is not available
			// Then, let everything go down in shame
		}
		return env;
	}

	/*
		The symbols in this context.
	*/
	public List symbols() {
		if (isEmpty()) return List.NIL;
		return cdr().symbols().cons(car().symbols());
	}

	/*
		Checks sym for being protected in this context.
	*/
	public void checkProtected(Symbol sym) throws Alarm {
		if (!this.protects(sym)) return;
		throw new Alarm("PROTECT", "Symbol " + sym + " is protected");
	}

	/*
		Checks sym for not being bound in this context.
	*/
	public void checkUnbound(Symbol sym) throws Alarm {
		Expr e = retrieve(sym);
		if (e == null) return;
		throw new Alarm("ISBOUND", "Symbol " + sym + " is already bound");
	}
}

//--------------------------------------------------------------------------------------------------
