
package expressions;

/*
	Frame.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	A Frame is a dictionary for bindings
	between symbols and expressions
*/

import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;

@SuppressWarnings("serial")
public class Frame extends HashMap<Symbol, Expr> {

	/*
		Protection table.
		(Initialization deferred to method protect)
	*/
	private HashSet<Symbol> ptable = null;

	/*
		Creates an empty frame.
	*/
	public Frame() {
		super();
	}

	/*
		Creates a frame from a list lsym of symbols
		and a list ls of values. 
		The caller must ensure that lsym is a symbol list
		and that lsym.length == ls.length.
	*/
	public Frame(List lsym, List ls) {
		// create a hash table of size 4
		super(4);
		while (!lsym.isEmpty()) {
			put((Symbol) lsym.car(), ls.car());
			lsym = lsym.cdr();
			ls = ls.cdr();
		}
	}

	/*
		Creates a frame from a list lsym of symbols.
		All values bound in the frame are void.
		The caller must ensure that lsym is a symbol list.
	*/
	public Frame(List lsym) {
		super(8);
		while (!lsym.isEmpty()) {
			put((Symbol) lsym.car(), VoidExpr.VOID);
			lsym = lsym.cdr();
		}
	}

	/*
		Stores a binding.
	*/
	public void bind(Symbol sym, Expr val) {
		put(sym, val);
	}

	/*
		Retrieves a value, or null if sym is unbound.
	*/
	public Expr retrieve(Symbol sym) {
		return get(sym);
	}

	/*
		Deletes a symbol.
	*/
	public void delete(Symbol sym) {
		remove(sym);
		if (ptable != null) ptable.remove(sym);
	}	

	/*
		Protects a symbol.
	*/
	public void protect(Symbol sym) {
		if (ptable == null) ptable = new HashSet<Symbol>(); 
		ptable.add(sym);
	}

	/*
		Returns true, if sym is protected.
	*/
	public boolean protects(Symbol sym) {
		return (ptable != null) ? ptable.contains(sym) : false;
	}

	/*
		String representation: Keys only.
		Ellipsis, if too large.
	*/
	private static int LARGE = 50;
	public String toString() {
		String items = (size() < LARGE) ?
			symbols().toStringWithout() : "..";
		return "{" + items + "}";
	}

	/*
		Access to the symbols bound in this frame.
	*/
	public List symbols() {
		Set<Symbol> keys = keySet();
		List ls = List.NIL;
		for (Symbol key : keys) ls = new List(key, ls);
		return ls;
	}

	/*
		Checks sym for being protected in this frame.
	*/
	public void checkProtected(Symbol sym) throws Alarm {
		if (!this.protects(sym)) return;
		throw new Alarm("PROTECT", "Symbol " + sym + " is protected");
	}
}

//--------------------------------------------------------------------------------------------------
