
package expressions;

/*
	Function.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	A Function is a user defined function
*/

import tanagra.Sys;

public class Function extends Procedure implements Traceable { 

	private static Symbol NONAME = Symbol.forName("function[]");

	/*
		Components
	*/
	protected List formals;              // Formal parameter list
	protected Expr body;                 // Body of the function
	protected Env context;               // The enclosed environment
	protected Symbol name;               // Internal function name
	protected boolean vararg;            // Variable argument number?
	protected boolean tracing;           // Tracing mode

	/*
		Access methods
	*/	
	public List formals() { return formals; }
	public Expr body()    { return body; }
	public Env context()  { return context; }
	public Symbol name()  { return name; }
	public Bool vararg()  { return Bool.valueOf(vararg); }

	/*
		A function with fixed argument number
		Caller must ensure that f is a symbol list.
	*/
	public Function(List f, Expr b, Env e, Symbol nm) {
		super();
		formals = f;
		body = b;
		context = e;
		name = nm;
		vararg = false;
		tracing = false;
	}

	/*
		A function with variable argument number
	*/
	public Function(Symbol s, Expr b, Env e, Symbol nm) {
		super();
		formals =  new List(s, List.NIL);
		body = b;
		context = e;
		name = nm;
		vararg = true;
		tracing = false;
	}

	/*
		Evaluation
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}	

	/*
		Apply the function to argument list args
	*/
	public Expr apply(List args, Env env, boolean tailPosition) throws Alarm {
		return apply(args, env, tailPosition, context);
	}

	/*
		Apply the function to argument list args
		Use context as the context for evaluating the body
	*/
	public Expr apply(List args, Env env, boolean tailPosition, Env context) throws Alarm {
		if (!vararg) {
			checkArity(args, formals.length());
		}
		Function fun = this;
		List values = args.evalItems(env);
		List valFirst = values;  // Save the argument values of the first call
		if (tracing()) traceCall(values, tailPosition);
		if (tailPosition) {
			// Uncomment the next line for more tracing information
			// if (tracing()) traceTailCall(values);
			throw new TailCall(fun, values, context);
		}
		while (true) {
			try {
				Expr e = fun.call(values, context);
				if (tracing()) traceReturn(e, valFirst);
				return e;
			} catch (TailCall tc) {
				fun     = tc.function;
				values  = tc.values;
				context = tc.context;
			}
		}
	}

	/*
		Call the function body using the context env
	*/
	public Expr call(List values, Env env) throws Alarm {
		if (vararg) {
			values = List.NIL.cons(values);
		}
		Env newenv = new Env(new Frame(formals, values), env);
		return body.eval(newenv, Sys.tailCallElimination());
	}

	/*
		Print calling information
	*/	
	private void traceCall(List args, boolean tailPos) {
		Sys.println("\tCall" + ((tailPos) ? "*" : " ") + "\t" + traced(args));
	}

	/*
		Print return information
	*/	
	private void traceReturn(Expr e, List args) {
		if (e instanceof Stream) e = new StringExpr("(" + ((Stream) e).car() + " ...)");
		Sys.println("\tReturn\t" + e + " from " + traced(args));
	}

	/*
		Print tail call quit information
	*/	
	private void traceTailCall(List val) {
		Sys.println("\tQuit*\t" + traced(val));
	}

	/*
		String form of args used for tracing
	*/
	private String traced(List args) {
		Symbol head = (name().toString().length() > 0) ? name() : NONAME;
		return args.cons(head).toString();
	}

	/*
		Runtime type
	*/
	public Expr type() {
		return Symbol.forName("Function");
	}

	/*
	        String representation
	*/
	public String toString() {
		return "function[" + name + "]";
	}

	/*
		Returns the tracing mode of this function
	*/
	public boolean tracing() {
		return tracing;
	}

	/*
		Toggles tracing mode between on and off.
	*/
	public boolean toggleTracingMode() {
		return tracing = !tracing;
	}
}	

//--------------------------------------------------------------------------------------------------
