
package expressions;

/*
	List.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	A List is an Expr and a list of Exprs.
*/

import tanagra.Sys;

public class List extends LinkList<Expr> implements Expr {

	/*
		true, if this is a symbol list
	*/	
	private final boolean symlist;

	/*
		Creates an empty list
	*/
	protected List() {
		super();
		symlist = true;
	}

	public final static List NIL = new List();

	/*
		Creates a nonempty list
	*/
	protected List(Expr e, List ls) {
		super(e, ls);
		symlist = (e instanceof Symbol) && ls.isSymList();
	}

	/*
		Access to symlist from Java
	*/
	public boolean isSymList() {
		return symlist;
	}

	/*
		Access to symlist from the t.Languages
	*/
	public Bool symlist() {
		return Bool.valueOf(symlist);
	}

	/*
		returns the extension of this list by expression e
	*/
	public List cons(Expr e) {
		return new List(e, this);
	}

	/*
		Access to head
	*/	
	public Expr car() {
		return head();
	}

	/*
		Typesafe access to tail
	*/	
	public List cdr() {
		return (List) tail();
	}

	/*
		Access to length
	*/	
	public Num len()  {
		return new Num(length);
	}

	/*
		Runtime type
	*/
	public Expr type() {
		return Symbol.forName("List");
	}

	/*
		Evaluation: First evaluate the car of this list.
		If a procedure is obtained, apply it to the cdr of this list.
		The tail position parameter is optional.
		If not given, it is set to false.
	*/
	public Expr eval(Env env, boolean... tailPosition) throws Alarm {
		Sys.checkCtrlC();                    // Stop if Ctrl-C was pressed
		checkEvalnil();
		EvalStack.push(this, env);
		boolean tail = (tailPosition.length != 0) ? tailPosition[0] : false;
		Expr res;
		try {
			Procedure proc = checkProcedure(car().eval(env));
			res = proc.apply(cdr(), env, tail);
		} catch (TailCall tc) {
			EvalStack.popNoProtocol();    // Don't show tail call returns
			throw tc;
		}
		EvalStack.pop();
		return res;
	}

	/*
		Evaluate the items in this list.
	*/
	public List evalItems(Env env) throws Alarm {
		if (isEmpty()) return this;
		Expr e = car().eval(env);   // evaluate car first!
		List ls = cdr().evalItems(env);
		return ls.cons(e);
	}	

	/*
		String form with special care for quote symbol
	*/
	public String toString() {
		boolean quoted = (length() == 2) &&
			car().equals(Symbol.forName("quote"));
		return quoted ? "'" + cdr().car() : toStringWith();
	}

	/*
		Check e for being a procedure
	*/
	private static Procedure checkProcedure(Expr e) throws Alarm {
		if (e instanceof Procedure) return (Procedure) e;
		throw new Alarm("HEAD", "List head " + e + " is not a procedure");
	}

	/*
		Check for attempting to evaluate the empty list
	*/
	private void checkEvalnil() throws Alarm {
		if (!this.isEmpty()) return;
		throw new Alarm("EVALNIL", "Attempt to evaluate an empty list");
	}

	/*
		Check e for being a list
	*/
	public static List checkList(Expr e, Expr client) throws Alarm {
		if (e instanceof List) return (List) e;
		throw new Alarm("TYPE", client + " expects a list argument");
	}

	/*
		Check e for being a nonempty list
	*/
	public static List checkNonemptylist(Expr e, Expr client) throws Alarm {
		List ls = checkList(e, client);
		if (!ls.isEmpty()) return ls;
		throw new Alarm("EMPTYLIST", client + " expects a nonempty list");
	}	

	/*
		Check e for being a symbol list
	*/
	public static List checkSymlist(Expr e, Expr client) throws Alarm {
		List ls = checkList(e, client);
		if (ls.isSymList()) return ls;
		throw new Alarm("SYMLIST", client + " expects a symbol list argument");
	}

	/*
		Check e for being a nonempty symbol list
	*/
	public static List checkNonemptysymlist(Expr e, Expr client) throws Alarm {
		List ls = checkSymlist(e, client);
		if (!ls.isEmpty()) return ls;
		throw new Alarm("NESLIST",
                        client + " expects a nonempty symbol list argument");
	}		
}

//--------------------------------------------------------------------------------------------------
