
package expressions;

/*
	Macro.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	Macro facility for the t.Languages
*/

import tanagra.Sys;

public class Macro extends Procedure implements Traceable {

	/*
		components
	*/	
	private List formals;            // Formal parameter list
	private Expr body;               // The body
	private Symbol name;             // Internal macro name
	private boolean vararg;          // Variable argument number?
	private boolean tracing;         // Tracing mode

	/*
		Access methods
	*/
	public List formals() { return formals; }
	public Expr body()    { return body; }
	public Symbol name()  { return name; }
	public Bool vararg()  { return Bool.valueOf(vararg); }

	/*
		A macro with fixed argument number
		Caller must ensure that f is a symbol list.
	*/
	public Macro(List f, Expr b, Symbol nm) {
		super();
		formals = f;
		body = b;
		name = nm;
		vararg = false;
		tracing = false;
	}

	/*
		A macro with variable argument number
	*/
	public Macro(Symbol s, Expr b, Symbol nm) {
		super();
		formals = new List(s, List.NIL);
		body = b;
		name = nm;
		vararg = true;
		tracing = false;
	}

	/*
		Evaluation
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}	

	/*
		Apply the macro to argument list args:
		Replace formals by unevaluated (!) arguments
		and evaluate the resulting expression
	*/
	public Expr apply(List args, Env env, boolean tailPosition) throws Alarm {
		if (vararg) {
			args = List.NIL.cons(args);
		} else {
			checkArity(args, formals.length());
		}
		Frame frame = new Frame(formals, args);
		Expr xbody = expand(body, frame);
		if (tracing()) trace(args, xbody);
		return xbody.eval(env, tailPosition);
	}

	/*
		Expand expression e using frame
	*/
	private static Expr expand(Expr e, Frame frame) {
		Expr res = null;
		if (e instanceof Symbol) {
			Expr val = frame.retrieve((Symbol) e);
			res = (val != null) ? val : e;
		} else if (e instanceof List) {
			res = expandItems((List) e, frame);
		} else {
			res = e;
		};
		return res;
	}

	/*
		Expand the items in a list using frame
	*/
	private static List expandItems(List ls, Frame frame) {
		List res = List.NIL;
		if (!ls.isEmpty()) {
			Expr car = expand(ls.car(), frame);
			List cdr = expandItems(ls.cdr(), frame);
			res = cdr.cons(car);
		};
		return res;
	}

	/*
		Runtime type
	*/
	public Expr type() {
		return Symbol.forName("Macro");
	}

	/*
		String representation
	*/
	public String toString() {
		return "macro[" + name + "]";
	}

	/*
		Print tracing information
	*/
	private void  trace(List values, Expr expandedBody) {
		Sys.println("Tracing macro " + name +
		          "\nParameters: " + formals +
		          "\n      |---> " + values +
		          "\nBody:       " + body + 
			  "\n      |---> " + expandedBody);
	}

	/*
		Returns the tracing mode of this macro
	*/
	public boolean tracing() {
		return tracing;
	}

	/*
		Toggles tracing mode between on and off
	*/
	public boolean toggleTracingMode() {
		return tracing = !tracing;
	}
}

//--------------------------------------------------------------------------------------------------
