
package expressions;

/*
	Num.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	Numbers for the t.Languages
*/

import java.math.BigDecimal;
import java.math.BigInteger;
import java.math.RoundingMode;

public class Num implements Expr {

	private final static BigDecimal ZERO = BigDecimal.ZERO;
	private final static BigDecimal ONE  = BigDecimal.ONE;

	private BigDecimal num;       // Numerator
	private BigDecimal den;       // Denominator

	/*
		Access to parts
	*/
	public Num num() { return new Num(num); }	
	public Num den() { return new Num(den); }	

	/*
		Constructor from a string (provided by the lexer)
	*/
	public Num(String n) {
		num = firstPart(n);
		den = secondPart(n);
	}

	/*
		Constructor from long
	*/
	public Num(long n) {
		this(new BigDecimal(n), ONE);
	}

	/*
		Constructor from a BigDecimal
	*/
	public Num(BigDecimal n) {
		this(n, ONE);
	}

	/*
		Constructor from two BigDecimals
		Assumes d != 0
	*/
	private Num(BigDecimal n, BigDecimal d) {
		int scale = Math.max(n.scale(), d.scale());
		if (scale > 0) {                          // decimal
			try {
				if (n.compareTo(ZERO) == 0) n = n.setScale(1);
				num = (d.compareTo(ONE) == 0) ? n : n.divide(d);
				den = ONE;
			} catch (ArithmeticException ae) {    // fraction
				num = n;
				den = d;
				expand();
			}
		} else {                                  // fraction
			if (d.compareTo(ZERO) < 0) {
				num = n.negate();
				den = d.negate();
			} else {
				num = n;
				den = d;
			}
			reduce();
		}
	}

	private static BigDecimal firstPart(String n) {
		return new BigDecimal(n.split("/")[0]);
	}

	private static BigDecimal secondPart(String n) {
		String[] a = n.split("/");
		if (a.length == 1) return ONE;    // no 2nd part
		BigDecimal d = new BigDecimal(a[1]);
		return d;
	}

	/*
		Strip trailing zeros from BigDecimal n
	*/
	private static BigDecimal stripZeros(BigDecimal n) {
		if (n.scale() > 1) {
			n = n.stripTrailingZeros();
			if (n.scale() <= 0) n = n.setScale(1);
		}
		return n;
	}

	/*
		Reduce this fraction
	*/
	private void reduce() {
		if ((num.compareTo(ONE) == 0) || (den.compareTo(ONE) == 0)) return;
		BigInteger n = num.toBigIntegerExact();
		BigInteger d = den.toBigIntegerExact();
		BigDecimal gcd = new BigDecimal(n.gcd(d));
		num = num.divide(gcd, 0, RoundingMode.UNNECESSARY);
		den = den.divide(gcd, 0, RoundingMode.UNNECESSARY);
	}

	/*
		Expand this fraction
	*/
	private void expand() {
		int scale = Math.max(num.scale(), den.scale());
		if (scale <= 0) return;
		num = num.scaleByPowerOfTen(scale);
		den = den.scaleByPowerOfTen(scale);
		reduce();
	}

	/*
		Runtime type
	*/
	public Expr type() {
		return Symbol.forName("Number");
	}
							
	/*
		Equality
	*/
	public boolean equals(Object e) {
		return (e instanceof Num) && (this.compareTo((Num) e) == 0);
	}

	/*
		Evaluation
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}

	/*
		String representation
	*/
	public String toString() {
		String s = num.toPlainString();
		return (isFraction()) ? s + "/" + den.toPlainString() : s;
	}

	/*
		Is this a fraction?
	*/
	public boolean isFraction() {
		return den.compareTo(ONE) != 0;
	}

	/*
		Is this an integer?
	*/
	public boolean isInteger() {
		return (den.compareTo(ONE) == 0) && (num.scale() <= 0);
	}

	/*
		Conversion to int. Information may get lost,
		cf. java.math.BigDecimal.intValue().
	*/
	public int toInt(Expr client) throws Alarm {
		try {
			return num.divide(den).intValueExact();
		} catch (ArithmeticException ae) {
			throw new Alarm("TOINT", this + " is too large for " + client); 
		}
	
	}

	/*
		Conversion to a fraction
	*/
	public Num toFraction() throws Alarm {
		Num n =  new Num(num, den);
		n.expand();
		return n;
	}

	/*
		Numerator
	*/
	public BigDecimal numerator() throws Alarm {
		if (isFraction() || isInteger()) return num;
		return toFraction().num;
	}	

	/*
		Denominator
	*/
	public BigDecimal denominator() throws Alarm {
		if (isFraction() || isInteger()) return den;
		return toFraction().den;
	}	

	/*
		Operations with numbers
	*/
	public Num negate() {
		return new Num(num.negate(), den);
	}

	public Num invert() throws Alarm {
		checkNonzero(num);
		return new Num(den, num);
	}

	private boolean isDecimal() {
		return den.compareTo(ONE) == 0;
	}

	public Num add(Num n) {
		if (isDecimal() && n.isDecimal()) {
			return new Num(num.add(n.num));		
		}
		BigDecimal a = num.multiply(n.den).add(n.num.multiply(den));
		BigDecimal b = den.multiply(n.den);
		return new Num(a, b);
	}

	public Num sub(Num n) {
		return add(n.negate());
	}

	public Num mult(Num n) {
		return new Num(num.multiply(n.num), den.multiply(n.den));
	}

	public Num div(Num n) throws Alarm {
		checkNonzero(n.num);
		return mult(n.invert());
	}

	public int compareTo(Num n) {
		if (isDecimal() && n.isDecimal()) {
			return num.compareTo(n.num);
		}	
		BigDecimal a = num.multiply(n.den);
		BigDecimal b = n.num.multiply(den);
		return a.compareTo(b);
	}

	public Num round(int n) {
		n = Math.max(0, n);
		return new Num(num.divide(den, n, RoundingMode.DOWN));
	}

	/*
		Check n for not being zero
	*/
	private static void checkNonzero(BigDecimal n) throws Alarm {
		// do not use n.equals(BigDecimal.ZERO) in the next line!
		if (n.compareTo(BigDecimal.ZERO) != 0) return;
		throw new Alarm("ZERODIV", "Division by zero");
	}

	/*
		Check e for being a number
	*/
	public static Num checkNum(Expr e, Expr client) throws Alarm {
		if (e instanceof Num) return ((Num) e);
		throw new Alarm("TYPE", client + " expects a number argument");
	}

	/*
		Check e for being a nonnegative integer
	*/
	public static int checkInt(Expr e, Expr client) throws Alarm {
		if ((e instanceof Num) && ((Num) e).isInteger()) {
			int n = ((Num) e).toInt(client);
			if (n < 0) {
				String mesg = client + " expects a nonnegative integer";
				throw new Alarm("NNEG", mesg);
			}
			return n;
		}	
		throw new Alarm("TYPE", client + " expects an integer argument");
	}
}

//--------------------------------------------------------------------------------------------------
