	
package expressions;

/*
	Operator.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	Operator is the base class for the t.Languages operators
*/

import java.util.Hashtable;
import java.util.Collection;

public abstract class Operator extends Procedure {

	protected static List NIL = List.NIL;
	protected static Symbol EMPTY = Symbol.forName("");

	/*
		Internal name
	*/
	private String name;

	/*
		Access to name
	*/	
	public Symbol name() { return Symbol.forName(name); }

	/*
		The operator table
	*/
	private static Hashtable<String, Operator> optable;

	/*
		Access to the operator table
	*/	
	public List optable() {
		Collection<Operator> operators = optable.values();
		List ls = List.NIL;
		for (Operator op : operators) ls = ls.cons(op);
		return ls;
	}

	/*
		Class constructor
		Operator table of initial size 100
	*/
	static {
		optable = new Hashtable<String, Operator>(100);
	}

	/*
		Constructor without argument, in order to allow
		dynamic creation via newInstance()
	*/
	protected Operator() {
		name = namepartFrom(getClass().getName());
	}

	/*
		Evaluation
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}	

	/*
		Application of an operator
		Overridden by the operators in TailCallOp.java
	*/
	public Expr apply(List args, Env env, boolean tailPosition) throws Alarm {
		return apply(args, env);
	}

	/*
		Application without regard to tail position.
		Overridden by all operators not in  TailCallOp.java.
		We could make this method abstract, but then it must
		be implemented in all subclasses - also in the operators
		in TailCallOp.java, where it is never used. 
	*/
	public Expr apply(List args, Env env) throws Alarm {
		throw new Alarm("APPLY", "Wrong call to " + this);
	}

	/*
		Runtime type
	*/
	public Expr type() {
		return Symbol.forName("Operator");
	}

	/*
		String representation
	*/
	public String toString() {
		return "op[" + name + "]";
	}

	/*
		Returns the operator to a name
	*/
	protected static Operator forName(String name) throws Alarm {
		Operator op = null;
		if (optable.containsKey(name)) {  // operator exists
			op = optable.get(name);
		} else {                          // create operator
			Class<Operator> opclass = checkOpclass(name);
			op = Operator.forClass(opclass);
			optable.put(name, op);
		}	
		return op;
	}

	/*
		Check string name for being the name of an operator class
	*/
	private static Class<Operator> checkOpclass(String name) throws Alarm {
		String classname = Operator.classnameFor(name);
		try {
			//return Class.forName(classname);
			@SuppressWarnings("unchecked")
			Class<Operator> result =
				(Class<Operator>)Class.forName(classname).asSubclass(Operator.class);
			return result;	

		}
		catch(ClassNotFoundException cnfe) {
			throw new Alarm("UNKNOWNOP", "Unknown operator " + name);
		}
		catch(NoClassDefFoundError ncfe) {
			throw new Alarm("UNKNOWNOP", "Unknown operator " + name);
		}
	}

	/*
		Returns an operator for class opclass
	*/
	private static Operator forClass(Class<Operator> opclass)
		throws Alarm {
		try {
			return opclass.newInstance();
		} catch (InstantiationException ie) {
		} catch (IllegalAccessException iae) {
		}	
		throw new Alarm("NEWOP",
			"Can't instantiate class " + opclass.getName());
	}

	/*
		Returns the operator class name "Operator_name"
	*/
	protected static String classnameFor(String name) {
		return "expressions.Operator_" + name;
	}

	/*
		Returns the name part from the classname "Operator_name"
	*/
	private static String namepartFrom(String classname) {
		int i = classname.lastIndexOf('_') + 1;
		return classname.substring(i);
	}


	/*
		Short forms for checks
	*/
	protected Symbol checkSymbol(Expr e) throws Alarm {
		return Symbol.checkSymbol(e, this);
	}

	protected List checkList(Expr e) throws Alarm {
		return List.checkList(e, this);
	}

	protected List checkNonemptylist(Expr e) throws Alarm {
		return List.checkNonemptylist(e, this);
	}

	protected List checkSymlist(Expr e) throws Alarm {
		return List.checkSymlist(e, this);
	}

	protected List checkNonemptysymlist(Expr e) throws Alarm {
		return List.checkNonemptysymlist(e, this);
	}

	protected StringExpr checkString(Expr e) throws Alarm {
		return StringExpr.checkStringexpr(e, this);
	}

	protected Num checkNum(Expr e) throws Alarm {
		return Num.checkNum(e, this);
	}

	protected int checkInt(Expr e) throws Alarm {
		return Num.checkInt(e, this);
	}
}

//--------------------------------------------------------------------------------------------------
