	
package expressions;

/*
	SpecialOp.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	The set of special operators for the t.Languages
*/

import tanagra.*;                            // Only tanagra.Token is not needed
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;

class Operator_clock extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 0);
		long clock = System.currentTimeMillis();
		return new Num(clock);
	}
}

class Operator_part extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 2);
		String name = checkSymbol(args.first().eval(env)).toString();
		return checkPart(name, args.second().eval(env));
	}

	//	Checks if part name of expression e exists.
	//	If so, it's value is returned.
	private static Expr checkPart(String name, Expr e) throws Alarm {
		Class<?> cls = e.getClass();
		try {	
			java.lang.reflect.Method m = cls.getMethod(name);
			Object res = m.invoke(e);
			if (res instanceof Expr) return (Expr) res;
		} catch (java.lang.reflect.InvocationTargetException ite) {
			Throwable t = ite.getCause();
			if (t instanceof Alarm) throw (Alarm) t;
		} catch (IllegalAccessException iae) {
		} catch (NoSuchMethodException nsme) {
		}
		throw new Alarm("PART", "No part '" + name + "' in " + e);
	}
}

class Operator_load extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 1);
		String file = checkString(args.first().eval(env)).toString();
		new Interpreter(file).run(env);
		return VoidExpr.VOID;
	}
}

class Operator_read extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 0);
		String s = new Lexer().getLine();	
		return new StringExpr((s == null) ? "" : s);  // Watch for empty input
	}
}

class Operator_parse extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 1);
		StringExpr s = checkString(args.first().eval(env));
		try {
			return new Parser(s).read();
		} catch (Alarm alarm) {
			if (alarm.isEof())
				throw new Alarm("PARSE", "Parsing empty input");
			throw alarm;
		}
	}
}

class Operator_trace extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		while (!args.isEmpty()) {
			Traceable traceable = checkTraceable(args.first().eval(env));
			boolean tracingMode = traceable.toggleTracingMode();
			Sys.println("Tracing mode for " + traceable + " is " +
				(tracingMode ? "on" : "off"));
			args = args.cdr();	
		}
		return VoidExpr.VOID;
	}

	//	Check e for being traceable
	private static Traceable checkTraceable(Expr e) throws Alarm {
		if (e instanceof Traceable) return (Traceable) e;
		throw new Alarm("TYPE", e + " cannot be traced");
	}
}

class Operator_protocol extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 0);
		boolean protocol = EvalStack.toggleProtocolMode();
		return new StringExpr("Protocol mode is " + (protocol ? "on" : "off"));
	}
}

class Operator_tail extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 0);
		boolean ttc = Sys.toggleTailCallElimination();
		return new StringExpr("Tail call elimination is " + (ttc ? "on" : "off"));
	}
}

class Operator_line extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 0);
		Sys.toggleLineNumbering();
		return VoidExpr.VOID;
	}
}

class Operator_exec extends Operator {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 1);
		String command = checkString(args.first().eval(env)).toString();
		Process proc = null;
		try {   // neuen Prozess erzeugen...
			proc = Runtime.getRuntime().exec(command);
		} catch(IOException ioe) {
			throw new Alarm("EXEC", "command not executable");
		}
		try {   // ...und dessen Input auf die eigene Ausgabe umlenken:
			BufferedReader rdr = new BufferedReader(
				new InputStreamReader(proc.getInputStream())
			);
			String s = rdr.readLine();
			if (s != null) Sys.println(s);
		} catch (IOException ioe) {
			throw new Alarm("EXEC", "command not executable");
		}
		return VoidExpr.VOID;
	}
}

//------------------------------------------------------------------------------
