
package expressions;

/*
	ClassObj.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	A ClassObj is a t.Java object representing a t.Java class.
*/

import tanagra.Interpreter;

public class ClassObj extends Obj implements Traceable {

	/*
		Components
	*/	
	private Symbol className;             // Name of this t.Java class
	private ClassObj superClass;          // Superclass of this class
	private List fields;                  // Fields of this class
	private Env methods;                  // Methods of this class
	private boolean sealed;               // Has this classObj been sealed?
	private boolean hasInit;              // Does this classObj have an init function?
	private boolean tracing;              // Tracing mode of this class

	/*
		Access methods
	*/
	public Symbol className()      { return className; }
	public Symbol name()           { return className; }  // alias, so that (name class) works
	public ClassObj superClass()   { return superClass; }
	public List fields()           { return fields; }
	public Env methods()           { return methods; }
	public boolean sealed()        { return sealed; }
	public static ClassObj CLASS() { return CLASS; }

	/*
		ClassObj CLASS is the class (= type) of a user defined class.
	*/
	private final static ClassObj CLASS;                        // The type 'Class'

	/*
		Class constructor of class ClassObj
	*/	
	static {
		CLASS = new ClassObj(Symbol.forName("Class"), List.NIL, Interpreter.globalenv());
		CLASS.classObj = CLASS;
		CLASS.sealed = true;
	}

	/*
		Incomplete private constructor
		superClass and methods are set in the public constructor.
	*/	
	private ClassObj(List attributes, Symbol name, Env env) {
		super(CLASS, env);
		className = name;
		superClass = null;
		fields = attributes;
		methods = null;
		sealed = false;
		hasInit = false;
		tracing = false;
	}

	/*
		Constructor for root classes
		Caller ensures that attributes is a symbol list.
	*/
	public ClassObj(Symbol name, List attributes, Env env) {
		this(attributes, name, env);
		superClass = this;
		methods = new Env(new Frame());
	}

	/*
		Constructor for derived classes
		Caller ensures that attributes is a symbol list.
	*/
	public ClassObj(Symbol name, ClassObj sclass, List attributes, Env env) {
		this(attributes, name, env);
		superClass = sclass;
		methods = new Env(new Frame(), sclass.methods);
	}

	/*
		Is this a root class?
	*/
	public boolean isRoot() {
		return this == superClass;
	}

	/*
		Does this t.Java class have an init method?
	*/	
	public boolean hasInitMethod() {
		if (isRoot()) {
			return hasInit;
		} else {	
			return hasInit || superClass.hasInitMethod();
		}	
	}

	/*
		The distance this class has from its root class
	*/
	public int depth() {
		return methods.length() - 1;
	}

	/*
		Find the method with descriptor m and n arguments.
		m must be a method name or a symbol list of length 2.
	*/
	public Method getMethod(Expr m, int n) throws Alarm {
		checkMethodcallsyntax(m);
		Symbol nm = null;           // Base name of the method
		ClassObj cls = null;        // Search starts here
		if (m instanceof Symbol) {
			nm = (Symbol) m;
			cls = this;
		} else {                    // m is a list of 2 symbols
			List ls = (List) m;
			Symbol className = (Symbol) ls.first();
			nm = (Symbol) ls.second();
			cls = getSuperclass(className);
		}
		Symbol name = nameWithUnderscore(nm, n);
		Method method = cls.retrieve(name);
		checkNotnull(method, name, cls);
		return method;
	}

	/*
		Check e for being a syntactically correct method call
	*/
	public static void checkMethodcallsyntax(Expr e) throws Alarm {
		if (e instanceof Symbol) return;
		if (e instanceof List) {
			List ls = (List) e;
			if (ls.isSymList() && (ls.length() == 2)) return;
		}
		throw new Alarm("MCALL",
			"Symbol or symbol list of length 2 expected as method descriptor");
	}


	/*
		Check method m for not being null
	*/
	private void checkNotnull(Method m, Symbol name, ClassObj cls) throws Alarm {
		if (m != null) return;
		String s = isRoot() ? "" : " or in a superclass";
		throw new Alarm("MSEARCH", "No method " + name + " in " + cls + s);
	}

	/*
		Retrieves the method according to a
		method name with underscore, like foo_3.
		Returns null, if not found
	*/	
	private Method retrieve(Symbol name) {
		return (Method) methods.retrieve(name);
	}

	/*
		Is there a Method str_0?
	*/
	private final static Symbol str_0 = str.withSuffix(0);
	public boolean hasFormatMethod() {
		return retrieve(str_0) != null;
	}

	/*
		Half method name, like foo_3
	*/
	private Symbol nameWithUnderscore(Symbol name, int n) {
		return Symbol.forName(name + "_" + n);
	}

	/*
		Installs a method for this type
		Caller ensures that formals is a symbol list
	*/	
	public Method putMethod(List formals, Expr body, Symbol sym)
							throws Alarm {
		checkSealed();	
		if (sym.equals(init)) hasInit = true;	
		Symbol name = nameWithUnderscore(sym, formals.length());
		Method m = new Method(this, formals, body, context, name);
		methods.put(name, m);
		return m;
	}

	/*
		Check for being a sealed class
	*/
	private void checkSealed() throws Alarm {
		if (!sealed()) return;
		throw new Alarm("SEALED", this + " is sealed");
	}


	/*
		Returns the superclass for a name, or throws an alarm
	*/	
	private ClassObj getSuperclass(Symbol name) throws Alarm {
		if (className == name) return this;
		if (isRoot()) throw new Alarm("FINDCLASS",
			"No superclass " + name + " in " + this);
		return superClass.getSuperclass(name);
	}

	/*
		Seales this class
	*/
	public void seal() {
		sealed = true;
	}

	/*
		String representation
	*/
	public String toString() {
		return "class[" + className + "]";
	}

	/*
		Generic name of an object of this class
	*/
	public String genericObjectName() {
		return "object[" + className + "]";
	}

	/*
		Returns the tracing mode of this class
	*/
	public boolean tracing() {
		return tracing;
	}

	/*
		Toggles tracing mode between on and off.
	*/
	public boolean toggleTracingMode() {
		return tracing = !tracing;
	}

	/*
		Check e for being a ClassObj
	*/
	public static ClassObj checkClassobj(Expr e, Expr client) throws Alarm {
		if (e instanceof ClassObj) return (ClassObj) e;
		throw new Alarm("TYPE", client + " expects a class object");
	}
}	

//--------------------------------------------------------------------------------------------------
