
package expressions;

/*
	Obj.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	An Obj is a t.Java object.
*/

import tanagra.Sys;

public class Obj extends Procedure {

	/*
		Components
	*/	
	protected ClassObj classObj;                  // t.Java class of this object
	protected Env      context;                   // Context of this object

	/*
		Access methods
	*/	
	public ClassObj classObj() { return classObj; }
	public Env      context()  { return context; }

	/*
		Constants
	*/	
	protected final static Symbol init = Symbol.forName("init");
	protected final static Symbol self = Symbol.forName("self");
	protected final static Symbol str  = Symbol.forName("str");

	/*
		Common constructor for ordinary and class objects
	*/
	protected Obj(ClassObj classobj, Env env) {
		classObj = classobj;
		context = env;
	}

	/*
		Constructor for non-class objects, with t.Java constructor call
	*/
	public Obj(ClassObj classobj, List args, Env env) throws Alarm {
		this(classobj, null);
		context = contextForClass(classobj);
		if ((args.length > 0) || classObj.hasInitMethod()) {
			apply(args.cons(init), env, false);
		}
	}

	/*
		Apply the object to its argument list
	*/
	public Expr apply(List args, Env env, boolean tailPosition) throws Alarm {
		checkAritymin(args, 1);                    // We need a method name
		int arity = args.length() - 1;             // Arity of the method call
		Method m = getMethod(args.first(), arity); // Method search
		Env bodycontext = contextForMethod(m);     // for evaluating the body of m
		return m.apply(args.cdr(), env, tailPosition, bodycontext);
	}	

	/*
		Method search (overridden in subclass ClassObj)
	*/
	protected Method getMethod(Expr m, int arity) throws Alarm {
		return classObj.getMethod(m, arity);
	}

	/*
		Returns a context for this object, starting with the fields of class classobj
	*/
	private Env contextForClass(ClassObj classobj) {
		Frame frame = new Frame(classobj.fields());
		if (classobj.isRoot()) {
			frame.put(self, this);
			// frame.protect(self);
			return new Env(frame, classobj.context());
		} else {
			return new Env(frame, contextForClass(classobj.superClass()));
		}
	}

	/*
		Get the context for method m
	*/
	private Env contextForMethod(Method m) {
		int delta = classObj.depth() - m.classObj().depth();
		Env env = context;
		for (int i = 0;  i < delta; i++) {
			env = env.cdr();
		}
		return env;
	}

	/*
		Runtime type
	*/
	public Expr type() {
		return classObj;
	}

	/*
		Evaluation
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}	

	/*
	        String representation
	*/
	public String toString() {
		return classObj.hasFormatMethod() ? format() :
			classObj.genericObjectName();
	}

	/*
		Formatted representation of this object
	*/	
	private String format() {
		try {
			return apply(List.NIL.cons(str), context, false).toString();
		} catch (Alarm alarm) {
			notice(alarm.content());
			return classObj.genericObjectName();
		}
	}

	/*
		Notice in case of error in format method
	*/	
	private void notice(Expr msg) {
		Sys.println("\tNon-fatal error \"" + msg + "\" occured while " +
		"formatting object.");
	}
}

//--------------------------------------------------------------------------------------------------
