	
package expressions;

/*
	ObjectOp.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	The base class for the t.Java operators
*/

import tanagra.Interpreter;

public abstract class ObjectOp extends Operator {

	/*
		Check e for being a ClassObj
	*/	
	protected ClassObj checkClassobj(Expr e) throws Alarm {
		return ClassObj.checkClassobj(e, this);
	}

	/*
		Check e for not being class[Class]
	*/	
	protected void checkNotCLASS(Expr e) throws Alarm {
		if (!e.equals(ClassObj.CLASS())) return;
		throw new Alarm("NEWCLASS", "Illegal use of " + this);
	}
}

/*
	The t.Java expression (new C x y ..) creates a new
	t.Java object obj of class C and calls (obj init x y ..).
*/
class Operator_new extends ObjectOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkAritymin(args, 1);
		ClassObj classObj = checkClassobj(args.first().eval(env));
		checkNotCLASS(classObj);  // No new classes with op[new].
		return new Obj(classObj, args.cdr(), env);
	}
}

/*
	A root class is created by calling (class name fields),
	or by calling  (class name) if it has no fields,
	a subclass of sclass is created by (class name sclass fields).
	The argument 'name' is not evaluated.
*/	
class Operator_class extends ObjectOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 1, 3);
		Symbol sym = checkSymbol(args.first());
		ClassObj sclass = null; // Superclass of this class
		List fields = null;
		Expr classObj = null;
		switch (args.length()) {
		case 1:	
			classObj = new ClassObj(sym, NIL, env);
			break;
		case 2:
			Expr arg2 = args.second().eval(env);
			if (arg2 instanceof ClassObj) {
				sclass = checkClassobj(args.second().eval(env));
				checkNotCLASS(sclass);
				classObj = new ClassObj(sym, sclass, NIL, sclass.context);
			} else {
				fields = checkSymlist(args.second().eval(env));
				classObj = new ClassObj(sym, fields, env);
			}
			break;
		case 3:
			sclass = checkClassobj(args.second().eval(env));
			checkNotCLASS(sclass);
			fields = checkSymlist(args.third().eval(env));
			classObj = new ClassObj(sym, sclass, fields, sclass.context);
			break;
		default :
			// Unreachable
		}
		return classObj;
	}
}

/*
	(method <class> <proto> <body>)
	adds a method to class <class>.
*/
class Operator_method extends ObjectOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 3);
		ClassObj classObj = checkClassobj(args.first().eval(env));
		List proto   = checkNonemptysymlist(args.second());
		Symbol name  = (Symbol) proto.car();
		List formals = proto.cdr();
		Expr body    = args.third();
		classObj.putMethod(formals, body, name);
		return VoidExpr.VOID;
	}
}

/*
	(seal A) forbids the addition of methods to class A
*/
class Operator_seal extends ObjectOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 1);
		ClassObj classObj = checkClassobj(args.first().eval(env));
		classObj.seal();
		return VoidExpr.VOID;
	}
}

//------------------------------------------------------------------------------
