
package expressions;

/*
	Record.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	A record is a compound expression
*/

public class Record extends Procedure {

	/*
		Contents
	*/
	private Expr[] element;
	private Symbol type;

	/*
		Create a record of type t and size n
	*/
	public Record(Symbol t, int n) {
		type = t;
		element = new Expr[(n < 0) ? 0 : n];
		for (int i = 0; i < size(); i++) {
			element[i] = VoidExpr.VOID;
		}
	}

	/*
		Elementwise equality
	*/
	public boolean equals(Object e) {
		if (!(e instanceof Record)) return false;
		Record r = (Record) e;
		if (size() != r.size() || type() != r.type()) return false;
		boolean res = true;
		for (int i = 0; res && (i < size()); i++) {
			res &= element[i].equals(r.element[i]);
		}
		return res;
	}

	/*
		String representation: elements in brackets
	*/
	public String toString() {
		String str = (size() == 0) ? "" : element[0].toString();
		for (int i = 1; i < size(); i++) {
			str += "," + element[i];
		}
		return "[" + str + "]";
	}

	/*
		Evaluation
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}

	/*
		Application
	*/
	public Expr apply(List args, Env env, boolean tailPosition) throws Alarm {
		checkArity(args, 0, 2);
		if (args.length() == 0) return new Num(size());
		int i = checkInt(args.first().eval(env));
		if (args.length() == 1) return elementAt(i);
		setElementAt(i, args.second().eval(env));
		return VoidExpr.VOID;
	}
	
	/*
		Runtime type
	*/
	public Expr type() {
		return type;
	}

	/*
		The element at index i
	*/
	private Expr elementAt(int i) throws Alarm {
		checkIndex(i, size());	
		return element[i];
	}

	/*
		At index i store expression e 
	*/
	private void setElementAt(int i, Expr e) throws Alarm {
		checkIndex(i, size());	
		element[i] = e;
	}

	/*
		Size
	*/
	private int size() {
		return element.length;
	}

	/*
		Check e for being an integer
	*/
	private int checkInt(Expr e) throws Alarm {
		return Num.checkInt(e, this.type());
	}

	/*
		Check i for being in [0..n)
	*/
	private void checkIndex(int i, int n) throws Alarm {
		if (0 <= i && i < n) return;
		String mesg = this.type() + " index " + i + " is out of bounds";
		throw new Alarm("INDEX", mesg);
	}

}

//--------------------------------------------------------------------------------------------------
