
package expressions;

/*
	Indeterminate.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	Indeterminates in t.Prolog expressions
*/

import java.util.Hashtable;

public class Indeterminate implements Expr {

	/*
		The classwide symbol table
	*/
	private final static Hashtable<Symbol, Indeterminate> table =
		new Hashtable<Symbol, Indeterminate>(100);

	/*
		The name of this indeterminate
	*/
	private Symbol name;

	/*
		Access to name
	*/
	public Symbol name() { return name; }

	/*
		Creates an indeterminate from a symbol
	*/
	private Indeterminate(Symbol s) {
		name = s;
	}

	/*
		Returns the indeterminate according to a symbol
	*/
	public static Indeterminate forSymbol(Symbol sym) {
		Indeterminate indet = table.get(sym);
		if (indet == null) {
			indet = new Indeterminate(sym);
			table.put(sym, indet);
		}
		return indet;
	}

	/*
		Creates a term from expression e. The term is a copy of e
		in which all symbols listed in symbols are replaced by indeterminates.
		Each occurence of the underscore symbol is replaced by a
		unique indeterminate.
		Symbols inside of records or t.Java objects are not replaced.
		Caller must ensure that argument symbols is a symbol list.
	*/
	private static Expr term(List symbols, Expr e) {
		if (e.equals(UNDERSCORE)) {
			return numberedIndeterminate();
		}
		if ((e instanceof Symbol) && symbols.contains(e))  {
			return Indeterminate.forSymbol((Symbol) e);
		}	
		if (e instanceof List) {
			List ls = (List) e;
			if (ls.isEmpty()) return ls;
			Expr head = term(symbols, ls.first());
			return ((List) term(symbols, ls.cdr())).cons(head);
		}	
		return e;
	}

	/*
		The underscore symbol
	*/	
	private final static Symbol UNDERSCORE = Symbol.forName("_");

	/*
		counter for numbered indeterminates
	*/	
	private static int N = 0;

	/*
		The next numbered indeterminate
	*/	
	private static Indeterminate numberedIndeterminate() {
		return Indeterminate.forSymbol(Symbol.forName("#" + ++N));
	}

	/*
		Wrapper for method term
	*/	
	public static Expr toTerm(List symbols, Expr e) {
		N = 0;
		return term(symbols, e);
	}

	/*
		Creates a copy of expression e, with all indeterminates
		appended by _n.
	*/
	public static Expr copy(Expr e, int n) {
		if (e instanceof Indeterminate) {
			Indeterminate x = (Indeterminate) e;
			String s = x.name.toString() + "_" + n;
			Symbol sym = Symbol.forName(s);
			return Indeterminate.forSymbol(sym);
		}	
		if (e instanceof List) {
			List ls = (List) e;
			if (ls.isEmpty()) return ls;
			Expr head = copy(ls.first(), n);
			return ((List) copy(ls.cdr(), n)).cons(head);
		}	
		return e;
	}

	/*
		Returns a copy of expression e, with all indeterminates
		X_n replaced by X.
	*/
	public static Expr uncopy(Expr e) {
		if (e instanceof Indeterminate) {
			Indeterminate x = (Indeterminate) e;
			String s = x.name.toString();
			int i = s.lastIndexOf('_');
			if (i == -1) return e;
			Symbol sym = Symbol.forName(s.substring(0, i));
			return Indeterminate.forSymbol(sym);
		}
		if (e instanceof List) {
			List ls = (List) e;
			if (ls.isEmpty()) return ls;
			Expr head = uncopy(ls.first());
			return ((List) uncopy(ls.cdr())).cons(head);
		}	
		return e;
	}

	/*
		Returns the substitution for the most general unifier
		of term s and term t if it exists, or null.
	*/
	public static Substitution unify(Expr s, Expr t) {
		if (s.equals(t)) {
			return new Substitution();
		}
		if (s instanceof Indeterminate) {
			Indeterminate x = (Indeterminate) s;
			if (x.occursIn(t)) return null;
			return new Substitution(x, t);  // x |-> t
		}
		if (t instanceof Indeterminate) {
			return unify(t, s);
		}	
		if ((s instanceof List) && (t instanceof List)) {
			return unifyLists((List) s, (List) t);
		}
		return null;
	}

	/*
		Returns the substitution for the most general unifier
		of (s1 ... sk) and (t1 ... tk) if it exists, or null.
	*/
	private static Substitution unifyLists(List s, List t) {
		if (s.length() != t.length()) {
			return null;
		}	
		if (s.isEmpty() && t.isEmpty()) {
			return new Substitution();
		}	
		Substitution alpha = unify(s.car(), t.car());
		if (alpha == null) return null;
		List ls = alpha.applyTo(s.cdr());
		List lt = alpha.applyTo(t.cdr());
		Substitution beta = unifyLists(ls, lt);
		if (beta == null) return null;
		return alpha.product(beta);
	}

	/*
		Returns true if this indeterminate occurs in term e
	*/	
	private boolean occursIn(Expr e) {
		if (e instanceof List) {
			List ls = (List) e;
			if (ls.isEmpty()) return false;
			return occursIn(ls.car()) || occursIn(ls.cdr());
		} else {
			return this.equals(e);
		}
	}

	/*
		t.Prolog type
	*/
	public Expr type() {
		return Symbol.forName("Indeterminate");
	}

	/*
		Evaluation
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}

	/*
		Equality
	*/
	public boolean equals(Object obj) {
		return this == obj;
	}

	/*
		String representation
	*/
	public String toString() {
		return name + "*";
	}
}

//--------------------------------------------------------------------------------------------------
