
package expressions;

/*
	PrologOp.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	Base class for the t.Prolog operators
*/
public abstract class PrologOp extends Operator {

	protected Substitution checkSubstitution(Expr e) throws Alarm {
		return Substitution.checkSubstitution(e, this);
	}

	protected Rule checkRule(Expr e) throws Alarm {
		return Rule.checkRule(e, this);
	}

	protected RuleBase checkRulebase(Expr e) throws Alarm {
		return RuleBase.checkRulebase(e, this);
	}
}

class Operator_term extends PrologOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 2);
		List varnames = checkSymlist(args.first());
		return Indeterminate.toTerm(varnames, args.second().eval(env));
	}
}

class Operator_unify extends PrologOp {
	private static Symbol fail = Symbol.forName("fail");
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 2);
		Expr e = Indeterminate.unify(args.first().eval(env),
						args.second().eval(env));
		return (e != null) ? e : fail;
	}
}

class Operator_subst extends PrologOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 2);
		Expr e = args.first().eval(env);
		Substitution sub = checkSubstitution(args.second().eval(env));
		return sub.applyTo(e);
	}
}

class Operator_product extends PrologOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 2);
		Substitution sub1 = checkSubstitution(args.first().eval(env));
		Substitution sub2 = checkSubstitution(args.second().eval(env));
		return sub1.product(sub2);
	}
}

class Operator_rulebase extends PrologOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 1);
		Symbol name = checkSymbol(args.first());
		return new RuleBase(name);
	}
}

class Operator_rule extends PrologOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 1);
		return new Rule(checkNonemptylist(args.first().eval(env)));
	}
}

class Operator_addrule extends PrologOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkArity(args, 2);
		RuleBase rulebase = checkRulebase(args.first().eval(env));
		Rule rule = checkRule(args.second().eval(env));
		rulebase.addRule(rule);
		return VoidExpr.VOID;
	}
}

class Operator_resolve extends PrologOp {
	public Expr apply(List args, Env env) throws Alarm {
		checkAritymin(args, 3);
		RuleBase rulebase = checkRulebase(args.first().eval(env));
		List symbols = checkSymlist(args.second());
		args = args.cdr().cdr();                // (arg3, arg4, ...)
		List goals = (List) Indeterminate.toTerm(symbols, args);
		boolean success = rulebase.resolve(goals, env);
		return Bool.valueOf(success);
	}
}

//------------------------------------------------------------------------------
