
package expressions;

/*
	Rule.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)
	
	Rules for t.Prolog
*/

public class Rule implements Expr {

	private final static Symbol ATOM = Symbol.forName("atom");

	/*
		Components of this rule
	*/
	private Expr   lhs;                          // left hand side
	private List   rhs;                          // right hand side
	private Symbol name;                         // name

	/*
		Access to components
	*/
	public Expr   lhs()  { return lhs; }  
	public List   rhs()  { return rhs; }
	public Symbol name() { return name; }

	/*
		Constructor
	*/
	private Rule(Expr e, List rs) {
		lhs =  e;
		rhs =  rs;
		name = nameFor(e);
	}

	/*
		Constructor from list.
		Caller ensures that ls is not empty.
	*/
	public Rule(List ls) throws Alarm  {
		this(ls.car(), ls.cdr());
	}

	/*
		Returns a name like f/2 for predicates like (f x y),
		list\n for non-predicate lists of length n,
		or else the symbol 'atom'.
	*/	
	public static Symbol nameFor(Expr e) {
		if (named(e)) {
			List pred = (List) e;
			Symbol sym = (Symbol) pred.car();
			int n = pred.length() - 1;
			return Symbol.forName(sym + "/" + n);
		} else if (e instanceof List) {
			int n = ((List) e).length();
			return Symbol.forName("list" + "/" + n);
		} else {
			return ATOM;
		}	
	}

	/*
		Returns true if e is a nonempty list whose head is a symbol.
	*/
	private static boolean named(Expr e) {
		if (!(e instanceof List)) return false;
		List ls = (List) e;
		return !ls.isEmpty() && (ls.car() instanceof Symbol);
	}

	/*
		Returns a copy of this rule, with
		all indeterminates appended by n.
	*/
	public Rule copy(int n) {
		Expr ls = Indeterminate.copy(lhs, n);
		List rs = (List) Indeterminate.copy(rhs, n); // safe!
		return new Rule(ls, rs);
	}

	/*
		Evaluation
	*/	
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}

	/*
		Runtime type
	*/
	public Expr type() {
		return Symbol.forName("Rule");
	}

	/*
		Equality is by value. Container classes need this to work
		properly for rules.
	*/
	public boolean equals(Object obj) {
		if (!(obj instanceof Rule)) return false;
		Rule rule = (Rule) obj;
		return name.equals(rule.name) &&
			lhs.equals(rule.lhs) &&
			rhs.equals(rule.rhs);
	}

	/*
		String form
	*/
	public String toString() {
		return lhs + (rhs.isEmpty() ? "" : " :- " + rhs.toStringWithout());
	}

	/*
		Check e for being a rule
	*/
	public static Rule checkRule(Expr e, Expr client) throws Alarm {
		if (e instanceof Rule) return (Rule) e;
		throw new Alarm("TYPE", client + " expects a rule");
	}
}

//--------------------------------------------------------------------------------------------------
