
package expressions;

/*
	RuleBase.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	Rule bases for t.Prolog
*/

import java.util.LinkedHashMap;
import java.util.ArrayList;
import tanagra.Interpreter;
import tanagra.Sys;

@SuppressWarnings("serial")
public class RuleBase implements Expr, Traceable {

	/*
		Constants
	*/	
	private final static Symbol ESC = Symbol.forName("!");
	private final static Symbol LIM = Symbol.forName("recursionLimit");
	private final static Symbol THR = Symbol.forName("loopThreshold");
	private final static int DEFAULT_LIM = 256;
	private final static List NIL = List.NIL;

	/*
		Fields
	*/
	private LinkedHashMap<Symbol, ArrayList<Rule>> store;         // Rule store
	private Symbol name;                     // Internal name of this rule base
	private boolean tracing;                                    // Tracing mode

	private int loopThreshold;                     // To stop (suspected) loops
	private int recursionLimit;                       // To stop deep recursion
	private boolean loopFound;                      // true if a loop was found
	private int cutLevel;                            // values > 0 signal a cut

	/*
		Access to name
	*/
	public Symbol name() { return name; }

	/*
		Creates an empty rule base.
	*/
	public RuleBase(Symbol nm) {
		super();
		store = new LinkedHashMap<Symbol, ArrayList<Rule>>();
		name = nm;
		tracing = false;
	}

	/*
		Adds a rule to this rule base.
	*/
	public void addRule(Rule rule) throws Alarm {
		Symbol name = rule.name();
		if (!store.containsKey(name)) {
			store.put(name, new ArrayList<Rule>());
		}
		ArrayList<Rule> rules = store.get(name);
		if (!rules.contains(rule)) {
			rules.add(rule);
		}
	}

	/*
		Returns the success of resolving a list of goals.
		Parameters:
			goals        - the terms to be resolved,
			env          - the context for the escape to t.Scheme.
	*/	
	public boolean resolve(List goals, Env env) throws Alarm {
		recursionLimit = getRecursionLimit();
		loopThreshold  = getLoopThreshold();
		cutLevel = 0;
		loopFound = false;
		boolean success = resolve(goals, 0, env, NIL); // May set loopFound to 'true'
		if (!success && loopFound) {
			throw new Alarm("LOOP",
				"An unrecoverable loop was found");
		}		
		return success;	
	}

	/*
		Tries to resolve a list of goals.
		Returns true if successful, else false.
		Parameters:
			goals - the terms to be proved,
			depth - the recursion depth of this resolution step,
			env   - a context for the escape to t.Scheme,
			open  - List of as yet unresolved goals, used for loop detection.
	*/
	private boolean resolve(List goals, int depth, Env env, List open)
								throws Alarm {
		if (goals.isEmpty()) {
			return true;                     // success
		}
		checkRecursionlimit(depth);              // Stop endless recursion
		Expr goal = goals.car();
		trace(goals.toStringWithout(), depth);
		if (escape(goal)) {                      // goal is (! ...)
			if (loopThreshold > 0)
				open = open.cons(Indeterminate.uncopy(goal));
			return resolveEscape(goals, depth + 1, env, open);
		}
		ArrayList<Rule> rules = rulesFor(goal);
		if (rules == null) {
			trace("- no rule", depth);
			return false;                    // failure
		}
		if (loopThreshold > 0) {
			open = open.cons(Indeterminate.uncopy(goal));
			if (loopDetected(open, depth)) {
				trace("- loop", depth);
				return false;            // failure
			}
		}
		for (Rule rule : rules) {
			if (cutLevel > depth) break;
			rule = rule.copy(depth);
			Expr head = rule.lhs();
			Substitution sub = Indeterminate.unify(goal, head);
			if (sub != null) {
				List newgoals = join(rule.rhs(), goals.cdr());
				newgoals = sub.applyTo(newgoals);
				if (resolve(newgoals, depth + 1, env, open)) {
					trace("+", depth);
					return true;     // success
				}
			}
		}
		trace("-", depth);
		return false;                            // failure
	}

	/*
		Tries to evaluate a list whose head is an escape goal.
		(Caller guarantees that the list has this form.)
		Returns true if successful, else false.
		Parameters are as for method resolve.
	*/
	private boolean resolveEscape(List goals, int depth, Env env, List open)
									throws Alarm {
		List esc = (List) goals.first();
		if (loopThreshold > 0) 
			open = open.cons(Indeterminate.uncopy(esc));
		int n = esc.length();
		if (n == 0 || n > 3)
			throw new Alarm("ESC", "Illegal escape goal " + esc);
		if (n == 1) {
			cutLevel = depth;
		}
		List newgoals = goals.cdr();
		if (n > 1) {
			Expr e = esc.second().eval(env);
			if (Bool.isFalse(e)) {
				trace("- from '!'", depth);
				return false;                     // failure
			} 
			if (n == 3) { 
				Expr arg3 = esc.third();
				Indeterminate x = checkIndeterminate(arg3);
				Substitution sub = new Substitution(x, e);
				newgoals = sub.applyTo(newgoals);
			}
		}	
		boolean success = resolve(newgoals, depth + 1, env, open);
		trace(success ? "+" : "-", depth);
		return success;
	}

	/*
		Check e for being an indeterminate
	*/
	private Indeterminate checkIndeterminate(Expr e) throws Alarm {
		if (e instanceof Indeterminate) return (Indeterminate) e;
		throw new Alarm("TYPE", "As second argument, ! expects an indeterminate");
	}

	/*
		Returns true if e has the form (! ...). 
	*/
	private boolean escape(Expr e) {
		if (!(e instanceof List)) return false;
		List ls = (List) e;
		return !ls.isEmpty() && ls.car().equals(ESC);
	}

	/*
		Returns true if a (suspected) loop was detected
		The list ls of terms must not be empty
	*/
	private boolean loopDetected(List ls, int depth) {
		Expr e = ls.car();
		ls = ls.cdr();
		int count = 0;
		while (!ls.isEmpty()) {
			if (ls.car().equals(e)) count++;
			if (count == loopThreshold) {
				loopFound = true;
				return true;
			}
			ls = ls.cdr();
		}
		return false;
	}

	/*
		Returns a user defined recursion limit,
		or else the value of DEFAULT_LIM.
	*/
	private int getRecursionLimit() throws Alarm {
		Expr e = Interpreter.globalenv().retrieve(LIM);
		if (e == null) return DEFAULT_LIM;
		return checkInt(e);
	}

	/*
		Returns a user defined loop threshold, or 0.
	*/
	private int getLoopThreshold() throws Alarm {
		Expr e = Interpreter.globalenv().retrieve(THR);
		if (e == null) return 0;
		return checkInt(e);
	}

	/*
		Returns the ArrayList with the rules for this goal,
		or null if there are no rules.
	*/
	private ArrayList<Rule> rulesFor(Expr goal) {
		return store.get(Rule.nameFor(goal));
	}

	/*
		Throws alarm if the recursion depth is too large
	*/
	private void checkRecursionlimit(int depth) throws Alarm {
		if (depth <= recursionLimit) return;
		throw new Alarm("RECLIM", "Recursion limit " +
						recursionLimit + " exceeded");
	}	

	/*
		Returns the join ls~rs of its argument lists.
		(Should rather be in class List.)
	*/	
	private static List join(List ls, List rs) {
		return ls.isEmpty() ? rs : join(ls.cdr(), rs).cons(ls.car());
	}

	/*
		If in tracing mode, print string s with indentation.
	*/	
	private void trace(String s, int depth) {
		if (!tracing()) return;
		String t = "    ";
		for (int i = 0; i < depth; i++) t += "| ";
		Sys.println(t + s); 
	}

	/*
		Evaluation
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}

	/*
		Runtime type
	*/
	public Expr type() {
		return Symbol.forName("Rulebase");
	}

	/*
		String form
	*/
	public String toString() {
		return "RuleBase[" + name + "]";
	}	

	/*
		Access to the content of this rule base
	*/
	public List content() {
		List ls = NIL;;
		for (ArrayList<Rule> rules : store.values()) {
			for (Rule rule : rules) ls = ls.cons(rule);
		}
		return ls;
	}

	/*
		Returns the tracing mode of this rule base
	*/
	public boolean tracing() {
		return tracing;
	}

	/*
		Toggles tracing mode between on and off.
	*/
	public boolean toggleTracingMode() {
		return tracing = !tracing;
	}

	/*
		Check e for being an integer
	*/	
	private int checkInt(Expr e) throws Alarm {
		return Num.checkInt(e, this);
	}

	/*
		Check e for being a rule base
	*/	
	public static RuleBase checkRulebase(Expr e, Expr client) throws Alarm {
		if (e instanceof RuleBase) return (RuleBase) e;
		throw new Alarm("TYPE", client + " expects a rule base");
	}
}	

//--------------------------------------------------------------------------------------------------
