
package expressions;

/*
	Substitution.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	Substitutions for t.Prolog.
	A substitution is a map from indeterminates to expressions.

*/

import java.util.HashMap;

@SuppressWarnings("serial")
public class Substitution extends HashMap<Indeterminate, Expr> implements Expr {

	/*
		Constructor for empty substitution
	*/
	public Substitution() {
		super();
	}

	/*
		Constructor for the substitution x -> e
	*/
	public Substitution(Indeterminate x, Expr e) {
		super();
		put(x, e);
	}

	/*
		Copy constructor
	*/
	private Substitution(Substitution s) {
		super(s);
	}

	/*
		Applies this substitution to expression e: All occurences in e
		of indeterminates in this substitution are replaced by their values.
		"All" is in our implementation not quite true: records, objects and
		other possible non-list compounds are not searched for occurences.
	*/
	public Expr applyTo(Expr e) {
		Expr res = null;
		if (e instanceof Indeterminate) {
			Indeterminate x = (Indeterminate) e;
			res = get(x);
			if (res == null) res = x;
		} else if (e instanceof List) {
			res = applyTo((List) e);
		} else if (e instanceof Substitution) {
			res = applyTo((Substitution) e);
		} else {
			res = e;
		}	
		return res;
	}

	/*
		 Applies this substitution to the elements of list ls.
	*/
	public List applyTo(List ls) {
		return ls.isEmpty() ? ls :
			applyTo(ls.cdr()).cons(applyTo(ls.car()));
	}

	/*
		 Applies this substitution to the
		 right hand sides of substitution s.
		 (Nondestructive, s remains unchanged.)
	*/
	private Substitution applyTo(Substitution s) {
		Substitution sub = new Substitution();
		for (Indeterminate x : s.keySet()) {
			Expr e = this.applyTo(s.get(x));
			if (!x.equals(e)) sub.put(x, e);
		}
		return sub;
	}

	/*
		The product of this substitution with substitution s
	*/
	public Substitution product(Substitution s) {
		Substitution sub = new Substitution();
		for (Indeterminate x : keySet()) {
			Expr e = s.applyTo(get(x));
			if (!x.equals(e)) sub.put(x, e);
		}
		for (Indeterminate x : s.keySet()) {
			if (!sub.containsKey(x)) sub.put(x, s.get(x));
		}
		return sub;
	}

	/*
		t.Prolog type
	*/
	public Expr type() {
		return Symbol.forName("Substitution");
	}

	/*
		Evaluation
	*/
	public Expr eval(Env env, boolean... tailPosition) {
		return this;
	}

	/*
		Check e for being a substitution
	*/
	public static Substitution checkSubstitution(Expr e, Expr client) throws Alarm {
		if (e instanceof Substitution) return (Substitution) e;
		throw new Alarm("TYPE", client + " expects a substitution");
	}
}

//--------------------------------------------------------------------------------------------------
