package tanagra;

/*
	Interpreter.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

	Interpreter for the t.Languages
*/

import expressions.*;

public class Interpreter {

	/*
		The global environment.
	*/	
	private static Env globalenv = Env.globalEnv();

	/*
		Access to the global environment.
	*/
	public static Env globalenv() {
		return globalenv;
	}

	/*
		The Parser delivers input expressions.
	*/	
	private Parser parser;

	/*
		An interpreter reading from stdin.
	*/	
	public Interpreter() {
		parser = new Parser();
	}

	/*
		An interpreter reading from a file.
	*/	
	public Interpreter(String filename) throws Alarm {
		parser = new Parser(filename);
	}

	/*
		Calls the main loop.
		Each argument must be a filename.
	*/
	public static void main(String[] args) {
		for (int i =  0; i <  args.length; i++) {
			try {
				new Interpreter(args[i]).run(globalenv);
			} catch (Alarm alarm) {
				Sys.println(alarm.toString());
			}
		};
		new Interpreter().run(globalenv);
	}

	/*
		Lets the interpreter run.
	*/
	public void run(Env env) {
		while (true) {
			try {
				read_eval_print(env);
			} catch (Alarm alarm) { // EOF, or an unforeseen exception
				if (alarm.isEof()) break;
				Sys.println("Unforeseen error.");
				break;
			} catch (StackOverflowError soe) {
				Sys.println("Stack overflow.");
				run(env);
				break;
			} catch (OutOfMemoryError ome) {
				Sys.println("Memory error.");
				run(env);
				break;
			}
		}
	}

	/*
		Reads, evaluates, and prints an expression.
	*/
	private void read_eval_print(Env env) throws Alarm {
		Expr expr = VoidExpr.VOID;          // input is void if parser fails
		Expr value = null;
		try {
			expr = parser.read();
			value = expr.eval(env, false);
		} catch (Alarm alarm) {
			if (alarm.isEof()) throw alarm; // Don't catch EOF
			value = alarm;
		} catch (StackOverflowError se) {
			value = new Alarm("STK", "Stack overflow");
		} catch (OutOfMemoryError oe) {
			value = new Alarm("MEM", "Not enough memory");
			Runtime.getRuntime().gc();  // Force garbage collection
		}
		EvalStack.reset();	
		store(expr);
		print(value);
	}

	/*
		Stores an expression.
	*/
	private static void store(Expr e) throws Alarm {
		int n = Sys.lineCounter();
		globalenv.bind(Symbol.forName("@" + n), e);
		globalenv.bind(Symbol.forName("@"), e);
	}

	/*
		Prints an expression.
	*/
	private static void print(Expr e) throws Alarm {
		String s = e.toString();
		int n = Sys.lineCounter();
		Sys.advanceLineCounter();
		globalenv.bind(Symbol.forName("%" + n), e);
		globalenv.bind(Symbol.forName("%"), e);
		if (!e.equals(VoidExpr.VOID)) Sys.println(s);
	}
}

//--------------------------------------------------------------------------------------------------
