package tanagra;

/*
	Lexer.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)

 	Lexer for the t.Languages
*/

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.FileReader;
import java.io.StringReader;
import java.io.IOException;
import expressions.Alarm;
import expressions.StringExpr;

public class Lexer {

	/*
	 	Special characters allowed in identifiers
	*/
	private final static String IDENT_CHAR = ".$/!%&=?^\\*+~@#<>|:,`-_";
	private final static String HOME_PREFIX = "+";
	private final static String HOME = System.getenv("TANAGRA");

	private BufferedReader source;    // Input source
	private String line;              // Current input line
	private int pos;                  // Position of current character
	private String sval;              // String value of the last token

	/*
		A lexer reading from stdin
	*/
	public Lexer() {
		InputStreamReader isr = new InputStreamReader(System.in);
		source = new BufferedReader(isr);
	}

	/*
		A lexer reading from a file
	*/
	public Lexer(String filename) throws Alarm {
		try {
			if (filename.length() == 0) throw new IOException();
			if (filename.startsWith(HOME_PREFIX)) {
				filename = HOME + filename.substring(1);
			}
			FileReader fr = new FileReader(filename);
			source = new BufferedReader(fr);
		} catch (IOException ioe) {
			String msg = "Unable to open file " + filename;
			throw new Alarm("NOFILE", msg);
		}
	}

	/*
		A lexer reading from a string expression
	*/
	public Lexer(StringExpr s) {
		StringReader srdr = new StringReader(s.toString());
		source = new BufferedReader(srdr);
	}

	/*
		The current input line
	*/
	public String currentLine() {
		return line;
	}
	
	/*
		Returns a new input line.
		Returns null, if EOF is reached.
	*/
	public String getLine() {
		try {
			line = source.readLine();
			pos = 0;
			sval = "";
		} catch (IOException ioe) {
			String msg = "I/O-error: " + ioe.getMessage();
			throw new RuntimeException(msg);
		}
		return line;
	}

	/*
		Returns the next token, changing sval and pos.
	*/
	public Token getToken() {
		sval = "";
		if (line == null) return Token.EOF;
		skipWhitespace();
		if (pos >= line.length()) return Token.EOL;
		Token tok = null;
		switch (chr()) {
		case '\'':
			nextPos();
			tok = Token.QUOTE;
			break;
		case '(':
			nextPos();
			tok = Token.LEFT;
			break;
		case ')':
			nextPos();
			tok = Token.RIGHT;
			break;
		case '\"' :
			tok = getString() ? Token.STRING : Token.USTRING;
			break;
		default:
			if (isNumberchar()) {
				getNumber();
				tok = infinity() ? Token.OTHER : Token.NUMBER;
			} else if (isIdentchar()) {
				getIdent();
				tok = Token.SYMBOL;
			} else {
				nextPos();
				tok = Token.OTHER;
			}
		}
		return tok;
	}

	/*
		True if sval is d...d/0...0
	*/
	private boolean infinity() {
		String[] a = sval.split("/");
		if (a.length == 1) return false; // no denominator
		return a[1].matches("0+");
	}

	/*
		Returns the string value of the token last read
	*/
	public String strVal() {
		return sval;
	}

	/*
		Read an identifier, assuming isNumberchar() == true.
	*/
	private void getIdent() {
		while (isIdentchar()) nextPos();
	}

	/*
		Read a number, assuming isNumberchar() == true.
	*/
	private void getNumber() {
		if (chr() == '-') nextPos();
		if (chr() == '+') pos++;         // Don't add leading '+' to sval
		while (isDigit(chr())) nextPos();
		if (chr() == '.') {
			nextPos();
		} else {	
			// Zahlen der Form d..d/d..d:
			if (chr() == '/' && isDigit(chr(1))) nextPos();
		}			 
		while (isDigit(chr())) nextPos();
	}

	/*
		Reads a string, assuming char() == '"'.
		Returns true, if the String is terminated by '"'.
	*/
	private boolean getString() {
		pos++;                            // don't add leading '"' to sval
		while (true) {
			switch (chr()) {
			case '"' :            // Don't add terminating '"' to sval
				pos++; 
				return true;
			case '\n' :                              // eol is reached
				return false;
			case '\\' :               // Backslash is escape character
				pos++;               // Skip the backslash
				switch (chr()) {
				case '"' :
					nextPos();
					break;
				case 'n' :
					sval += '\n'; pos++;
					break;
				case 't':
					sval += '\t'; pos++;
					break;
				case '\\':
					sval += '\\'; pos++;
					break;
				default :
					sval += '\\';       // Don't increment pos
				}
				break;
			default :
				nextPos();
			}
		}
	}

	/*
		Skip whitespace and comments
	*/
	private void skipWhitespace() {
		if (line == null) return;
		while((chr() != '\n') && Character.isWhitespace(chr())) pos++;
		if (chr() == ';') {                            // skip comment
			while (chr() != '\n') pos++;
		}
	}

	/*
		Returns the current character
	*/
	private char chr() {
		return chr(0);
	}

	/*
		Returns the character at pos + k
	*/
	private char chr(int k) {
		return (pos + k >= line.length()) ? '\n' : line.charAt(pos + k);
	}

	/*
		Advance to the next input position
	*/
	private void nextPos() {
		sval += chr();
		pos++;
	}

	/*
		Is c a digit?
	*/
	private boolean isDigit(char c) {
		return Character.isDigit(c);
	}

	/*
		Is chr() the beginning of a number?
	*/
	private boolean isNumberchar() {
		if (isDigit(chr())) return true;
		switch (chr()) {
		case '-' :
		case '+' :
			return isDigit(chr(1)) || (chr(1)=='.' && isDigit(chr(2)));
		case '.' :
			return isDigit(chr(1));
		default:
			return false;
		}
	}

	/*
		Is chr() an identifier character?
	*/
	private boolean isIdentchar() {
		return IDENT_CHAR.indexOf(chr()) >= 0 ||
			Character.isLetterOrDigit(chr());
	}
}

//--------------------------------------------------------------------------------------------------
