package tanagra;

/*
	Parser.java
	Copyright (c) A. Clausing (achim.clausing@uni-muenster.de)
	Parser for the t.Languages
*/

import expressions.*;

public class Parser {
	
	private static List NIL = List.NIL;

	private Lexer lexer;               // The lexer to read from
	private Token token;               // The most recent token
	private boolean interactive;       // Is this parser used interactively?

	/*
		A parser reading from stdin
	*/
	public Parser() {
		lexer = new Lexer();
		interactive = true;
		Sys.printPrimaryPrompt();
		getLine();
	}

	/*
		A parser reading from a file
	*/
	public Parser(String filename) throws Alarm {
		lexer = new Lexer(filename);
		interactive = false;
		getLine();
	}

	/*
		A parser reading from a string
	*/
	public Parser(StringExpr s) throws Alarm {
		lexer = new Lexer(s);
		interactive = false;
		getLine();
	}

	/*
		Returns an input expression
	*/
	public Expr read() throws Alarm {
		if (interactive) {
			while (token == Token.EOL) {
				Sys.printPrimaryPrompt();
				getLine();
			}	
		}
		Expr expr = parseExpr();
		if (token != Token.EOL) {
			while (token != Token.EOL) getToken();
			throw new Alarm("TRAILING", "Trailing input");
		}	
		return expr;
	}

	/*
		Parses an expression
	*/
	private Expr parseExpr() throws Alarm {
		Expr e = null;
		switch (token) {
		case EOL :
			if (interactive) Sys.printSecondaryPrompt();
			getLine();
			e = parseExpr();
			break;
		case EOF :
			throw new Alarm("EOF", "Unexpected end of input");
		case NUMBER :
			Num n = new Num(lexer.strVal());
			getToken();
			e = n;
			break;
		case SYMBOL :
			e = Symbol.forName(lexer.strVal());
			getToken();
			break;
		case STRING :
			e = new StringExpr(lexer.strVal());
			getToken();
			break;
		case USTRING :
			getToken();
			throw new Alarm("USTRING", "Unfinished string");
		case QUOTE :
			getToken();
			e = quote(parseExpr());
			break;
		case LEFT :
			getToken();
			e = parseList();
			break;
		case RIGHT :
			getToken();
			throw new Alarm("RIGHT", "Expression cannot begin with ')'");
		case OTHER :
			while (token != Token.EOL) getToken();
			throw new Alarm("ILLEGAL", "Illegal token");
		default :
			// unreachable
		}
		return e;
	}

	/*
		Parses a list of expressions
	*/
	private List parseList() throws Alarm {
		List ls = null;
		switch (token) {
		case EOL :
			if (interactive) Sys.printSecondaryPrompt();
			getLine();
			ls = parseList();
			break;
		case EOF :
			throw new Alarm("ULIST", "Unterminated list in input");
		case RIGHT :
			getToken();
			ls = NIL;
			break;
		default :
			Expr car = parseExpr();
			List cdr = parseList();
			ls = cdr.cons(car);
		}
		return ls;
	}

	/*
		Returns the same as '(quote e)'
	*/
	private Expr quote(Expr e) {
		return  NIL.cons(e).cons(Symbol.forName("quote"));
	}

	/*
		Get a new input line for this parser.
	*/
	private void getLine() {
		lexer.getLine();
		getToken();
	}

	/*
		Read a new token to variable 'token'
	*/
	private void getToken() {
		token = lexer.getToken();
	}
}

//--------------------------------------------------------------------------------------------------
